package com.gxzc.zen.common.config.cache.redis

import com.fasterxml.jackson.annotation.JsonTypeInfo
import com.fasterxml.jackson.core.JsonGenerator
import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.databind.SerializerProvider
import com.fasterxml.jackson.databind.module.SimpleModule
import com.fasterxml.jackson.databind.ser.std.StdSerializer
import org.apache.commons.pool2.impl.GenericObjectPool
import org.springframework.beans.factory.ObjectProvider
import org.springframework.beans.factory.annotation.Qualifier
import org.springframework.boot.autoconfigure.condition.ConditionalOnClass
import org.springframework.boot.autoconfigure.condition.ConditionalOnMissingBean
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty
import org.springframework.boot.autoconfigure.data.redis.RedisProperties
import org.springframework.boot.context.properties.EnableConfigurationProperties
import org.springframework.cache.CacheManager
import org.springframework.cache.annotation.EnableCaching
import org.springframework.cache.support.NullValue
import org.springframework.context.annotation.Bean
import org.springframework.context.annotation.Configuration
import org.springframework.context.annotation.Primary
import org.springframework.data.redis.cache.RedisCacheManager
import org.springframework.data.redis.connection.RedisClusterConfiguration
import org.springframework.data.redis.connection.RedisConnectionFactory
import org.springframework.data.redis.connection.RedisNode
import org.springframework.data.redis.connection.RedisSentinelConfiguration
import org.springframework.data.redis.connection.jedis.JedisConnectionFactory
import org.springframework.data.redis.core.RedisTemplate
import org.springframework.data.redis.core.StringRedisTemplate
import org.springframework.data.redis.serializer.GenericJackson2JsonRedisSerializer
import org.springframework.data.redis.serializer.StringRedisSerializer
import org.springframework.util.Assert
import org.springframework.util.StringUtils
import redis.clients.jedis.JedisPoolConfig
import java.io.IOException
import java.net.URI
import java.net.URISyntaxException
import java.text.SimpleDateFormat
import java.util.*

/**
 * Redis 缓存配置
 * 缓存管理器
 * @author NorthLan
 * @date 2018/3/29
 * @url https://noahlan.com
 */
@Configuration
@EnableCaching
@ConditionalOnProperty(prefix = "spring.cache", name = ["type"], havingValue = "redis")
@EnableConfigurationProperties(RedisProperties::class)
class RedisConfig {

    @Bean("redisCacheManager")
    @Primary
    fun cacheManager(@Qualifier("redisTemplate") redisTemplate: RedisTemplate<Any, Any>): CacheManager {
        return RedisCacheManager(redisTemplate.apply {
            val stringRedisSerializer = StringRedisSerializer()
            keySerializer = stringRedisSerializer
            hashKeySerializer = stringRedisSerializer
            // mapper
            val classPropertyTypeName: String? = null
            valueSerializer = GenericJackson2JsonRedisSerializer(ObjectMapper().apply {
                dateFormat = SimpleDateFormat("yyyy-MM-dd HH:mm:ss")
                registerModule(SimpleModule().addSerializer(NullValueSerializer(classPropertyTypeName)))
                if (StringUtils.hasText(classPropertyTypeName)) {
                    enableDefaultTypingAsProperty(ObjectMapper.DefaultTyping.NON_FINAL, classPropertyTypeName)
                } else {
                    enableDefaultTyping(ObjectMapper.DefaultTyping.NON_FINAL, JsonTypeInfo.As.PROPERTY)
                }
            })
        })
    }

    /**
     * [StdSerializer] adding class information required by default typing. This allows de-/serialization of
     * [NullValue].
     *
     * @author Christoph Strobl
     * @since 1.8
     */
    private class NullValueSerializer : StdSerializer<NullValue> {
        companion object {
            private const val serialVersionUID = 1999052150548658809L
        }

        private val classIdentifier: String

        /**
         * @param classIdentifier can be null and will be defaulted to `@class`.
         */
        constructor(classIdentifier: String?) : super(NullValue::class.java) {
            this.classIdentifier = if (StringUtils.hasText(classIdentifier)) classIdentifier!! else "@class"
        }

        /*
		 * (non-Javadoc)
		 * @see com.fasterxml.jackson.databind.ser.std.StdSerializer#serialize(java.lang.Object, com.fasterxml.jackson.core.JsonGenerator, com.fasterxml.jackson.databind.SerializerProvider)
		 */
        @Throws(IOException::class)
        override fun serialize(value: NullValue, jgen: JsonGenerator, provider: SerializerProvider) {
            jgen.writeStartObject()
            jgen.writeStringField(classIdentifier, NullValue::class.java.name)
            jgen.writeEndObject()
        }
    }

    /**
     * Redis connection configuration.
     */
    @Configuration
    @ConditionalOnClass(GenericObjectPool::class)
    protected class RedisConnectionConfiguration(private val properties: RedisProperties,
                                                 sentinelConfiguration: ObjectProvider<RedisSentinelConfiguration>,
                                                 clusterConfiguration: ObjectProvider<RedisClusterConfiguration>) {

        private val sentinelConfiguration: RedisSentinelConfiguration? = sentinelConfiguration.ifAvailable

        private val clusterConfiguration: RedisClusterConfiguration? = clusterConfiguration.ifAvailable

        protected val sentinelConfig: RedisSentinelConfiguration?
            get() {
                if (this.sentinelConfiguration != null) {
                    return this.sentinelConfiguration
                }
                val sentinelProperties = this.properties.sentinel
                if (sentinelProperties != null) {
                    val config = RedisSentinelConfiguration()
                    config.master(sentinelProperties.master)
                    config.setSentinels(createSentinels(sentinelProperties))
                    return config
                }
                return null
            }

        @Bean("redisConnectionFactory")
        @Primary
//        @ConditionalOnMissingBean(RedisConnectionFactory::class)
        fun redisConnectionFactory(): JedisConnectionFactory {
            return applyProperties(createJedisConnectionFactory())
        }

        protected fun applyProperties(
                factory: JedisConnectionFactory): JedisConnectionFactory {
            configureConnection(factory)
            if (this.properties.isSsl) {
                factory.isUseSsl = true
            }
            factory.database = this.properties.database
            if (this.properties.timeout > 0) {
                factory.timeout = this.properties.timeout
            }
            return factory
        }

        private fun configureConnection(factory: JedisConnectionFactory) {
            if (StringUtils.hasText(this.properties.url)) {
                configureConnectionFromUrl(factory)
            } else {
                factory.hostName = this.properties.host
                factory.port = this.properties.port
                if (this.properties.password != null) {
                    factory.password = this.properties.password
                }
            }
        }

        private fun configureConnectionFromUrl(factory: JedisConnectionFactory) {
            val url = this.properties.url
            if (url.startsWith("rediss://")) {
                factory.isUseSsl = true
            }
            try {
                val uri = URI(url)
                factory.hostName = uri.host
                factory.port = uri.port
                if (uri.userInfo != null) {
                    var password = uri.userInfo
                    val index = password.indexOf(":")
                    if (index >= 0) {
                        password = password.substring(index + 1)
                    }
                    factory.password = password
                }
            } catch (ex: URISyntaxException) {
                throw IllegalArgumentException("Malformed 'spring.redis.url' $url",
                        ex)
            }

        }

        /**
         * Create a [RedisClusterConfiguration] if necessary.
         * @return null if no cluster settings are set.
         */
        protected fun getClusterConfiguration(): RedisClusterConfiguration? {
            if (this.clusterConfiguration != null) {
                return this.clusterConfiguration
            }
            if (this.properties.cluster == null) {
                return null
            }
            val clusterProperties = this.properties.cluster
            val config = RedisClusterConfiguration(
                    clusterProperties.nodes)

            if (clusterProperties.maxRedirects != null) {
                config.maxRedirects = clusterProperties.maxRedirects!!
            }
            return config
        }

        private fun createSentinels(sentinel: RedisProperties.Sentinel): List<RedisNode> {
            val nodes = ArrayList<RedisNode>()
            for (node in StringUtils
                    .commaDelimitedListToStringArray(sentinel.nodes)) {
                try {
                    val parts = StringUtils.split(node, ":")
                    Assert.state(parts.size == 2, "Must be defined as 'host:port'")
                    nodes.add(RedisNode(parts[0], Integer.valueOf(parts[1])))
                } catch (ex: RuntimeException) {
                    throw IllegalStateException(
                            "Invalid redis sentinel property '$node'", ex)
                }

            }
            return nodes
        }

        private fun createJedisConnectionFactory(): JedisConnectionFactory {
            val poolConfig = if (this.properties.pool != null)
                jedisPoolConfig()
            else
                JedisPoolConfig()

            if (sentinelConfig != null) {
                return JedisConnectionFactory(sentinelConfig, poolConfig)
            }
            return if (getClusterConfiguration() != null) {
                JedisConnectionFactory(getClusterConfiguration(), poolConfig)
            } else JedisConnectionFactory(poolConfig)
        }

        private fun jedisPoolConfig(): JedisPoolConfig {
            val config = JedisPoolConfig()
            val props = this.properties.pool
            config.maxTotal = props.maxActive
            config.maxIdle = props.maxIdle
            config.minIdle = props.minIdle
            config.maxWaitMillis = props.maxWait.toLong()
            return config
        }

    }

    @Configuration
    class RedisConfiguration {
        @Bean
        @ConditionalOnMissingBean(name = ["redisTemplate"])
        fun redisTemplate(@Qualifier("redisConnectionFactory") redisConnectionFactory: RedisConnectionFactory): RedisTemplate<Any, Any> {
            val template = RedisTemplate<Any, Any>()
            template.connectionFactory = redisConnectionFactory
            return template
        }

        @Bean
        @ConditionalOnMissingBean(StringRedisTemplate::class)
        fun stringRedisTemplate(@Qualifier("redisConnectionFactory") redisConnectionFactory: RedisConnectionFactory): StringRedisTemplate {
            val template = StringRedisTemplate()
            template.connectionFactory = redisConnectionFactory
            return template
        }
    }
}